#!/usr/bin/env python
'''
Created on Feb 27, 2014

@author: k.edeline
'''

import numpy as np
import statsmodels.api as sm
import matplotlib.pyplot as plt
import matplotlib.mlab as mlab
import matplotlib
import operator, math
from collections import Counter
from mpl_toolkits.mplot3d import Axes3D

import Misc

# sources dict, replace by your owns
_V4_MONITORS={ 
               '91.200.16.102':("ams2-nl","Europe"),
               '84.88.81.122': ("bcn-es","Asia")
             }
_V6_MONITORS={ 
               '2001:40b0:7500:14::5458:517a': ("bcn-es","Asia"), 
               '2001:67c:1a8:200::102': ("ams2-nl","Europe")
              }

class Plotbox(object):
    '''
    A tracebox plotting class.
    '''

    def __init__(self, parsebox, output_folder=None):
        '''
        @param parsebox the Parsebox object
        '''
        self.parsebox=parsebox
        self.output_folder=output_folder

        self.monitors_v4=_V4_MONITORS
        self.monitors_v6=_V6_MONITORS
        matplotlib.rcParams.update({'font.family'   : 'serif', 'font.size'   : 17.0})
        
    def __path_length_list(self,arr):
        l=[]
        for dst in arr:
            for src in arr[dst]:
                average=0.0
                counter=0
                for tb in arr[dst][src]:
                    if tb.status == "success":
                        
                        leng=tb.path_length()
                        if leng > 2.0:
                            counter+=1
                            average+=leng
                if counter > 0:
                    l.append(average/counter)
        return l
    
    def __mb_refine(self,quarter,threshold=1.0):
        """
        return histo array with rare modifications merged
        keys for values under threshold are merged under 'Others' key

        @param threshold The percentage of modifications under which to
               merge into 'Others' category.
        """
        refined={}
        for modif, count in quarter.iteritems():
            if count<threshold:
                lab="Others"
                if lab in refined:
                    refined[lab]+=count
                else:
                    refined[lab]=count
            else:
                refined[modif]=count
    
        return refined
    
    def __cmp_lists(self,a,b):
        if len(a) != len(b) or len(a)<1:
            return len(a)-len(b)
        
        diff=0
        for e1 in a:
            if e1 not in b:
                diff+=1
                
        return diff
        
    def as_path_cmp(self):
        """
        Plots cdfs of the levenshtein distances between v4/v6 AS paths
        """
        t="total"
        diff_tab_best={t:[]}   
        diff_tab_worst={t:[]}   
        total_counter6=0
        cok=0
        cko=0
        for dst in self.parsebox.array6:
            for src in self.parsebox.array6[dst]:
                if src == "::1":
                    continue
                for tb in self.parsebox.array6[dst][src]:
                    if tb.status != "success":
                        continue
                    total_counter6+=1
                    asp6=None
                    asp4=None
                    asp6 = tb.get_asn_list(self.parsebox)
                    dst4=self.parsebox.get_v4_from_v6(dst)
                    src4=self.parsebox.get_monitor_v4_from_v6(src)

                    if dst4 in self.parsebox.array4:
                        if src4 in self.parsebox.array4[dst4]:
                            for tb in self.parsebox.array4[dst4][src4]:
                                if tb.status != "success":
                                    continue
                                asp4 = tb.get_asn_list(self.parsebox)
                                break
                    
                    continent=self.ip_to_continent(src, 6)
                    
                    if asp4 != None and asp6 != None:
                        if asp4 ==[] or asp6 ==[]:
                            cko+=1
                        else:
                            worst=Misc.levenshtein(asp4, asp6)
                            best=Misc.levenshtein(asp4, asp6,lower=False)
                            diff_tab_best[t].append(best)
                            if continent in diff_tab_best:
                                diff_tab_best[continent].append(best)
                            else:
                                diff_tab_best[continent]=[best]
                                
                            diff_tab_worst[t].append(worst)
                            if continent in diff_tab_worst:
                                diff_tab_worst[continent].append(worst)
                            else:
                                diff_tab_worst[continent]=[worst]
                            
                            cok+=1
                    else:
                        cko+=1       
        # Processing data per continent
        print "Per-continent v4/v6 levenshtein distances means and std dev:"
        print
        for continent in diff_tab_best:
            print continent
            l=diff_tab_worst[continent]
            sample = np.array(l)
            print "Worst case (unknown ASs never matches)"
            print sample.mean()
            print sample.std()
            ecdf = sm.distributions.ECDF(sample)
            xmin=min(sample)
            xmax=max(sample)
            x = np.linspace(xmin, xmax,num=30)
            y = ecdf(x)

            l=diff_tab_best[continent]
            sample = np.array(l)
            print "Best case (unknown AS always matches)"
            print sample.mean()
            print sample.std()
            print
            ecdf = sm.distributions.ECDF(sample)
            x2 = np.linspace(min(sample), max(sample),num=30)
            y2 = ecdf(x2)
            
            fig = plt.figure()
            ax = fig.add_subplot(111)
            ax.set_xlabel('v4/v6 AS Paths Levenshtein distance')
            ax.set_ylabel('Cumulative distribution function')
            for i in range(0,30,5):
                if i>=xmax:
                    break
                if i>=xmin:
                    ax.plot([i, i],[0, 1], linewidth=2, linestyle='dotted', color='0.7')

            ax.step(x, y, label="Worst case",color='r')
            ax.step(x2, y2, label="Best case",color='g')
            ax.legend(loc=4)
            plt.tight_layout()
            
            if self.output_folder == None:
                plt.show()
            else:
                plt.savefig(self.output_folder+'asn_leven_cmp_'+continent+'_cdf.pdf')    
                
            plt.close(fig)
            del fig               
  
    def as_path_len_cmp(self):
        """
        Plots v4/v6 as-path length comparison cdf and prints per-continet means
        """
        t="total"
        diff_tab={t:[]}     
        total_counter6=0
        cok=0
        cko=0
        for dst in self.parsebox.array6:
            for src in self.parsebox.array6[dst]:
                if src == "::1":
                    continue
                for tb in self.parsebox.array6[dst][src]:
                    if tb.status != "success":
                        continue
                    total_counter6+=1
                    len6=None
                    #print tb.get_asn_set(self.parsebox)
                    len6 = float(len(tb.get_asn_set(self.parsebox)))
                    dst4=self.parsebox.get_v4_from_v6(dst)
                    src4=self.parsebox.get_monitor_v4_from_v6(src)
                    len4=None
                    
                    if dst4 in self.parsebox.array4:
                        if src4 in self.parsebox.array4[dst4]:
                            for tb in self.parsebox.array4[dst4][src4]:
                                if tb.status != "success":
                                    continue
                                len4 = float(len(tb.get_asn_set(self.parsebox)))
                                break
                    
                    continent=self.ip_to_continent(src, 6)
                    if len4 != None and len6 != None:
                        if len4<2.0 or len6 < 2.0:
                            val=0.0
                            cko+=1
                        else:
                            val=(len4-len6)/len4
                            if val<0.5 and val>-0.5:
                                
                                diff_tab[t].append(val)
                            
                                if continent in diff_tab:
                                    diff_tab[continent].append(val)
                                else:
                                    diff_tab[continent]=[val]
                                cok+=1
                    else:
                        cko+=1       
        print "Per-continent number of crossed AS comparison means (refers to plots)."
        for continent in diff_tab:
            l=diff_tab[continent]
            sample = np.array(l)
            print continent
            print sample.mean()
            ecdf = sm.distributions.ECDF(sample)
            xmin=min(sample)
            xmax=max(sample)
            x = np.linspace(xmin, xmax)
            y = ecdf(x)
                    
            fig = plt.figure()
            ax = fig.add_subplot(111)
            ax.set_xlabel('number of crossed ASs relative v4/v6 difference')
            ax.set_ylabel('Cumulative distribution function (CDF)')
            ax.plot([0, 0],[0, 1], linewidth=2, linestyle='dotted', color='0.7')
            ax.annotate("IPv4 crossed\n less ASs",(xmin*.75,.5))
            ax.annotate("IPv4 crossed\n more ASs",(xmax*.25,.5))
            ax.step(x, y)
            ax.legend()
    
            if self.output_folder == None:
                plt.show()
            else:
                plt.savefig(self.output_folder+'asn_cmp_'+continent+'_cdf.pdf')    
                
            plt.close(fig)
            del fig           
    
    def middleboxes_modif_categories(self,top=5,show_quarters=True,close_to_vp=3,core=10):
        """
        Print statistics in <plot-dir>/modifs.log

        """    
        t="total"
        # IPv4
        src_tab4={t:[]}
        total_counter4=0
        for dst in self.parsebox.array4:
            for src in self.parsebox.array4[dst]:
                if "127.0.0.1" in src:
                    continue
                for tb in self.parsebox.array4[dst][src]:
                    if tb.status != "success":
                        continue
                    total_counter4+=1
                    modif,size = tb.modif_origin()
                    continent=self.ip_to_continent(src, 4)
                    src_tab4[t].append((modif,size))
                    if continent in src_tab4:
                        src_tab4[continent].append((modif,size))
                    else:
                        src_tab4[continent]=[(modif,size)]      
        # IPv6   
        src_tab6={t:[]}
        total_counter6=0
        for dst in self.parsebox.array6:
            for src in self.parsebox.array6[dst]:
                if src == "::1":
                    continue
                for tb in self.parsebox.array6[dst][src]:
                    if tb.status != "success":
                        continue
                    total_counter6+=1
                    modif,size = tb.modif_origin()
                    continent=self.ip_to_continent(src, 6)
                    src_tab6[t].append((modif,size))
                    if continent in src_tab6:
                        src_tab6[continent].append((modif,size))
                    else:
                        src_tab6[continent]=[(modif,size)]
        # Output to file
        f=open(self.output_folder+"modifs.log",'w')
        f.write("\nIPv4\n")        
        means4={}
        for continent in src_tab4:
            counters4=[0,0,0]
            quarter1_4={}
            quarter2_4={}
            quarter3_4={}
            complete={}
            for m,s in src_tab4[continent]:
                multi_modif={}
                for mm,hop in  m.iteritems():
                    if hop in multi_modif:
                        multi_modif[hop].append(mm)
                    else:
                        multi_modif[hop]=[mm]
                
                for hop,mods in multi_modif.iteritems():
                    mods=tuple(mods)
                    
                    found=0
                    for k in complete.keys():
                        if not self.__cmp_lists(tuple(k), mods):
                            complete[tuple(k)]+=1.0
                            found=1
                            break
                    if not found:
                        complete[mods]=1.0
                        
                    found=0
                    if hop<=close_to_vp:
                        for k in quarter1_4.keys():
                            if not self.__cmp_lists(tuple(k), mods):
                                quarter1_4[tuple(k)]+=1.0
                                found=1
                                break
                        if not found:
                            quarter1_4[mods]=1.0
                            
                        counters4[0]+=1
                    elif hop<=core:
                        for k in quarter2_4.keys():
                            if not self.__cmp_lists(tuple(k), mods):
                                quarter2_4[tuple(k)]+=1.0
                                found=1
                                break
                        if not found:
                            quarter2_4[mods]=1.0
                            
                        counters4[1]+=1
                    else:
                        for k in quarter3_4.keys():
                            if not self.__cmp_lists(tuple(k), mods):
                                quarter3_4[tuple(k)]+=1.0
                                found=1
                                break
                        if not found:
                            quarter3_4[mods]=1.0
                            
                        counters4[2]+=1
            
            f.write("\nContinent: %s"%continent)
            if show_quarters:
                f.write("\nClose to VP\n")
                i=0
                sorted_x = sorted(quarter1_4.iteritems(), key=operator.itemgetter(1),reverse=True)
                for modifs, v in sorted_x:
                    f.write("%s: %d times\n"%(modifs,v))
                    if i>=top:
                        break
                    i+=1      
                f.write("\ncore\n")
                i=0
                sorted_x = sorted(quarter2_4.iteritems(), key=operator.itemgetter(1),reverse=True)
                for modifs, v in sorted_x:
                    f.write("%s: %d times\n"%(modifs,v))
                    if i>=top:
                        break
                    i+=1
                f.write("\nClose to DST\n")
                i=0
                sorted_x = sorted(quarter3_4.iteritems(), key=operator.itemgetter(1),reverse=True)
                for modifs, v in sorted_x:
                    f.write("%s: %d times\n"%(modifs,v))
                    if i>=top:
                        break
                    i+=1
            f.write("\nOver the entire path\n")
            i=0
            sorted_x = sorted(complete.iteritems(), key=operator.itemgetter(1),reverse=True)
            for modifs, v in sorted_x:
                f.write("%s: %d times\n"%(modifs,v))

                i+=1     
        f.write("\nIPv6\n")
        means6={}
        for continent in src_tab6:
            counters6=[0,0,0]
            quarter1_6={}
            quarter2_6={}
            quarter3_6={}
            complete={}
            for m,s in src_tab6[continent]:
                multi_modif={}
                for mm,hop in  m.iteritems():
                    if hop in multi_modif:
                        multi_modif[hop].append(mm)
                    else:
                        multi_modif[hop]=[mm]
                
                for hop,mods in multi_modif.iteritems():
                    mods=tuple(mods)
                    
                    found=0
                    for k in complete.keys():
                        if not self.__cmp_lists(tuple(k), mods):
                            complete[tuple(k)]+=1.0
                            found=1
                            break
                    if not found:
                        complete[mods]=1.0
                    
                    found=0
                    if hop<=close_to_vp:
                        for k in quarter1_6.keys():
                            if not self.__cmp_lists(tuple(k), mods):
                                quarter1_6[tuple(k)]+=1.0
                                found=1
                                break
                        if not found:
                            quarter1_6[mods]=1.0
                            
                        counters6[0]+=1
                    elif hop<=core:
                        for k in quarter2_6.keys():
                            if not self.__cmp_lists(tuple(k), mods):
                                quarter2_6[tuple(k)]+=1.0
                                found=1
                                break
                        if not found:
                            quarter2_6[mods]=1.0
                            
                        counters6[1]+=1
                    else:
                        for k in quarter3_6.keys():
                            if not self.__cmp_lists(tuple(k), mods):
                                quarter3_6[tuple(k)]+=1.0
                                found=1
                                break
                        if not found:
                            quarter3_6[mods]=1.0
                            
                        counters6[2]+=1
            
            f.write("\nContinent: %s\n"%continent)
            if show_quarters:
                f.write("\nclose to VP\n")
                i=0
                sorted_x = sorted(quarter1_6.iteritems(), key=operator.itemgetter(1),reverse=True)
                for modifs, v in sorted_x:
                    f.write("%s: %d times\n"%(modifs,v))
                    if i>=top:
                        break
                    i+=1      
                f.write("\ncore\n")
                i=0
                sorted_x = sorted(quarter2_6.iteritems(), key=operator.itemgetter(1),reverse=True)
                for modifs, v in sorted_x:
                    f.write("%s: %d times\n"%(modifs,v))
                    if i>=top:
                        break
                    i+=1
                f.write("\nclose to dst\n")
                i=0
                sorted_x = sorted(quarter3_6.iteritems(), key=operator.itemgetter(1),reverse=True)
                for modifs, v in sorted_x:
                    f.write("%s: %d times\n"%(modifs,v))
                    if i>=top:
                        break
                    i+=1
            f.write("\nOver the entire path\n")
            i=0
            sorted_x = sorted(complete.iteritems(), key=operator.itemgetter(1),reverse=True)
            for modifs, v in sorted_x:
                f.write("%s: %d times\n"%(modifs,v))

                i+=1     
        f.close()
        print "Statistics outputed to <plot-dir>/modifs.log"
    
    def middleboxes_continent(self):
        """
        Plots middlebox modifications breakdown per-continent.
        """
        src_tab4={}
        total_counter4=0
        for dst in self.parsebox.array4:
            for src in self.parsebox.array4[dst]:
                if "127.0.0.1" in src:
                    continue
                for tb in self.parsebox.array4[dst][src]:
                    if tb.status != "success":
                        continue
                    total_counter4+=1
                    modif,size = tb.modif_origin()
                    continent=self.ip_to_continent(src, 4)
                    if continent in src_tab4:
                        src_tab4[continent].append((modif,size))
                    else:
                        src_tab4[continent]=[(modif,size)]
                        
        src_tab6={}
        total_counter6=0
        for dst in self.parsebox.array6:
            for src in self.parsebox.array6[dst]:
                if src == "::1":
                    continue
                for tb in self.parsebox.array6[dst][src]:
                    if tb.status != "success":
                        continue
                    total_counter6+=1
                    modif,size = tb.modif_origin()
                    continent=self.ip_to_continent(src, 6)
                    if continent in src_tab6:
                        src_tab6[continent].append((modif,size))
                    else:
                        src_tab6[continent]=[(modif,size)]
        means4={}
        for continent in src_tab4:
            counters4=[0,0,0]
            quarter1_4={}
            quarter2_4={}
            quarter3_4={}
            for m,s in src_tab4[continent]:
                for modif, hop in m.iteritems():
                    if hop<=3:
                        if modif in quarter1_4:
                            quarter1_4[modif]+=1.0
                        else:
                            quarter1_4[modif]=1.0
                        counters4[0]+=1
                    elif hop<=10:
                        if modif in quarter2_4:
                            quarter2_4[modif]+=1.0
                        else:
                            quarter2_4[modif]=1.0
                        counters4[1]+=1
                    else:
                        if modif in quarter3_4:
                            quarter3_4[modif]+=1.0
                        else:
                            quarter3_4[modif]=1.0
                        counters4[2]+=1
            
            quarter1_4=self.__mb_refine(quarter1_4)
            quarter2_4=self.__mb_refine(quarter2_4)
            quarter3_4=self.__mb_refine(quarter3_4)
            keys4=list(set(quarter1_4.keys()+quarter2_4.keys()+quarter3_4.keys()))
            total4=sum(counters4)
            mean4={}
            for k in keys4:
                mean4[k] = [round(quarter1_4[k]/total4,5)] if k in quarter1_4 else [0.0]
                mean4[k].append(round(quarter2_4[k]/total4,5) if k in quarter2_4 else 0.0)
                mean4[k].append(round(quarter3_4[k]/total4,5) if k in quarter3_4 else 0.0)
            N4=3.0 # nb monitor
            cm=plt.get_cmap("Paired")
            label4=["Close to VP", "core","Close to Dst"]
            ind = np.arange(N4)  # the x locations for the groups
            width = 1.0/len(keys4)     # the width of the bars

            rects=[]
            fig, ax = plt.subplots()
            i=0
            for s in mean4:
                rects.append(ax.bar(ind+i*width, tuple(mean4[s]), width, color=cm(1.*i/len(keys4)),label=s))
                i+=1
            
            # add text
            ax.set_ylabel('Rate')
            ax.set_title('IPv4 commons middleboxes modifications')
            ax.legend(prop={'size':8}, loc=3,bbox_to_anchor=(1, 0),fancybox=True, shadow=True)
    
            ax.set_xticks(ind)
            ax.set_xticklabels(tuple(label4),rotation=40,ha='right')
            for v in ind:
                ax.plot([ind, ind],[0, 1], linewidth=2, linestyle='dotted', color='0.7')
            
            plt.yscale('log', nonposy='clip')
            plt.tight_layout()
            plt.subplots_adjust(right=0.7)
            plt.savefig(self.output_folder+'mb_quarters_'+continent+'_v4.pdf')
            means4[continent]=mean4.copy()
            plt.close(fig)
            del fig
           
        # IPv6
        means6={}
        for continent in src_tab6: 
            counters6=[0,0,0]
            quarter1_6={}
            quarter2_6={}
            quarter3_6={}
            for m,s in src_tab6[continent]:
                for modif, hop in m.iteritems():
                    if hop<=3:
                        if modif in quarter1_6:
                            quarter1_6[modif]+=1.0
                        else:
                            quarter1_6[modif]=1.0
                        counters6[0]+=1
                    elif hop<=9:
                        if modif in quarter2_6:
                            quarter2_6[modif]+=1.0
                        else:
                            quarter2_6[modif]=1.0
                        counters6[1]+=1
                    else:
                        if modif in quarter3_6:
                            quarter3_6[modif]+=1.0
                        else:
                            quarter3_6[modif]=1.0
                        counters6[2]+=1
            
            quarter1_6=self.__mb_refine(quarter1_6)
            quarter2_6=self.__mb_refine(quarter2_6)
            quarter3_6=self.__mb_refine(quarter3_6)
            keys6=list(set(quarter1_6.keys()+quarter2_6.keys()+quarter3_6.keys()))
            total6=sum(counters6)
            mean6={}
            for k in keys6:
                mean6[k] = [round(quarter1_6[k]/total6,5)] if k in quarter1_6 else [0.0]
                mean6[k].append(round(quarter2_6[k]/total6,5) if k in quarter2_6 else 0.0)
                mean6[k].append(round(quarter3_6[k]/total6,5) if k in quarter3_6 else 0.0)  
            cm=plt.get_cmap("Paired")
            N6=3.0
            width2 = 1.0/len(keys6)
            rects2=[]
            label4=["Close to VP", "core","Close to Dst"]
            fig2, ax2 = plt.subplots()
            i=0
            for s in mean6:
                rects2.append(ax2.bar(ind+i*width2, tuple(mean6[s]), width2, color=cm(1.*i/len(keys6)),label=s))
                i+=1
            
            # add text
            ax2.set_ylabel('Rate')
            ax2.set_title('IPv6 commons middleboxes modifications')
            ax2.legend(prop={'size':8}, loc=3,bbox_to_anchor=(1, 0),fancybox=True, shadow=True)
    
            ax2.set_xticks(ind)
            ax2.set_xticklabels(tuple(label4),rotation=40,ha='right')
            for v in ind:
                ax2.plot([ind, ind],[0, 1], linewidth=2, linestyle='dotted', color='0.7')
            
            plt.yscale('log', nonposy='clip')
            plt.tight_layout()
            plt.subplots_adjust(right=0.65)
            plt.savefig(self.output_folder+'mb_quarters_'+continent+'_v6.pdf')            
            means6[continent]=mean6.copy()
            plt.close(fig2)
            del fig2
        
        # 3D per continent plots
        fig3 = plt.figure()
        ax3 = fig3.add_subplot(111, projection='3d')
        scaler=1.0
        cardin4=[]
        for cont in means4:
            cardin4+=means4[cont].keys()
            for s, rates in means4[cont].iteritems():
                for r in rates:
                    if r >0.0:
                        tmpscale=1.0
                        while r*tmpscale <1.0:
                            tmpscale*=10.0
                        scaler=max(scaler,tmpscale)
                             
        cardin4=list(set(cardin4))
        
        zt=[]
        zl=[]
        for z, continent in zip([0,10,20,30], means4.keys()):
            width3 = len(cardin4)
            tmpx=np.arange(width3)
            tmpy=np.arange(len(tmpx))
            i=0
            col_leg=[]
            lab_leg=[]
            for s in cardin4:  
                this_color=cm(1.*i/width3)
                col_leg.append(matplotlib.figure.Rectangle((0, 0), 1, 1, fc=this_color))
                lab_leg.append(s)          
                ax3.bar(i+ind*width3, tuple([math.log(x*scaler) if x>0.0 else 0.0 for x in means4[continent][s]]) if s in means4[continent] else (0.0, 0.0, 0.0), zs=z, zdir='y', color=this_color, alpha=0.8,label=s)
                i+=1
            zt.append(z)
            zl.append(continent)

        ax3.set_xticks(np.arange(len(ind)*width3))
        ll=[""]*width3*len(ind)
        labi=0
        for lab in label4:
            ll[labi]=lab
            labi+=width3
        ax3.set_xticklabels(tuple(ll),rotation=40,ha='right')
        ax3.set_yticks(zt)
        ax3.set_yticklabels(zl)
        ax3.set_ylabel("Continent")
        ax3.set_zlabel('Rate')
        ax3.zaxis.set_ticklabels([])
        ax3.set_title('IPv4 commons middleboxes modifications')
        ax3.autoscale(tight=True)
        ax3.legend(col_leg,lab_leg,prop={'size':8}, loc=3,bbox_to_anchor=(1, 0),fancybox=True, shadow=True)

        plt.subplots_adjust(right=0.7)
        plt.savefig(self.output_folder+'mb_quarters_3d_v4.pdf')  
        plt.close(fig3)
        del fig3
        
        fig4 = plt.figure()
        ax4 = fig4.add_subplot(111, projection='3d')
        scaler=1.0
        cardin6=[]
        for cont in means6:
            cardin6+=means6[cont].keys()
            for s, rates in means6[cont].iteritems():
                for r in rates:
                    if r >0.0:
                        tmpscale=1.0
                        while r*tmpscale <1.0:
                            tmpscale*=10.0
                        scaler=max(scaler,tmpscale)
                             
        cardin6=list(set(cardin6))
        zt=[]
        zl=[]

        for z, continent in zip([0,10,20,30], means6.keys()):
            width4 = len(cardin6)
            col_leg=[]
            lab_leg=[]
            i=0
            for s in cardin6:  
                this_color=cm(1.*i/width4)
                col_leg.append(matplotlib.figure.Rectangle((0, 0), 1, 1, fc=this_color))
                lab_leg.append(s)
                ax4.bar(i+ind*width4, tuple([math.log(x*scaler) if x>0.0 else 0.0 for x in means6[continent][s]]) if s in means6[continent] else (0.0, 0.0, 0.0), zs=z, zdir='y', color=this_color, alpha=0.8,label=s)
                i+=1
                
            zt.append(z)
            zl.append(continent)

        ax4.set_xticks(np.arange(len(ind)*width4))
        ll=[""]*width4*len(ind)
        labi=0
        for lab in label4:
            ll[labi]=lab
            labi+=width4
        ax4.set_xticklabels(tuple(ll),rotation=40,ha='right')
        ax4.set_yticks(zt)
        ax4.set_yticklabels(zl)
        ax4.set_ylabel("Continent")
        ax4.set_zlabel('Rate')
        ax4.zaxis.set_ticklabels([])
        ax4.set_title('IPv6 commons middleboxes modifications')
        ax4.autoscale(tight=True)
        ax4.legend(col_leg,lab_leg,prop={'size':8}, loc=3,bbox_to_anchor=(1, 0),fancybox=True, shadow=True)
    
        plt.subplots_adjust(right=0.7)
        plt.savefig(self.output_folder+'mb_quarters_3d_v6.pdf')  
        plt.close(fig4)
        del fig4

        print "Plots outputed to <plot-dir>/modifs.log"
    
    def middleboxes_overall(self):
        """
        Plots middlebox modifications breakdown.
        """
        t="total"
        src_tab4={t:[]}
        total_counter4=0
        for dst in self.parsebox.array4:
            for src in self.parsebox.array4[dst]:
                if "127.0.0.1" in src:
                    continue
                for tb in [self.parsebox.array4[dst][src][0]]:
                    if tb.status != "success":
                        continue
                    total_counter4+=1
                    modif,size = tb.modif_origin()
                    src_tab4[t].append((modif,size))
                    if src in src_tab4:
                        src_tab4[src].append((modif,size))
                    else:
                        src_tab4[src]=[(modif,size)]
                        
        src_tab6={t:[]}
        total_counter6=0
        for dst in self.parsebox.array6:
            for src in self.parsebox.array6[dst]:
                if src == "::1":
                    continue
                for tb in [self.parsebox.array6[dst][src][0]]:
                    if tb.status != "success":
                        continue
                    total_counter6+=1
                    modif,size = tb.modif_origin()
                    src_tab6[t].append((modif,size))
                    if src in src_tab6:
                        src_tab6[src].append((modif,size))
                    else:
                        src_tab6[src]=[(modif,size)]
          
        counters4=[0,0,0]
        quarter1_4={}
        quarter2_4={}
        quarter3_4={}
        for m,s in src_tab4[t]:
            for modif, hop in m.iteritems():
                if hop<=3:
                    if modif in quarter1_4:
                        quarter1_4[modif]+=1.0
                    else:
                        quarter1_4[modif]=1.0
                    counters4[0]+=1
                elif hop <=10:
                    if modif in quarter2_4:
                        quarter2_4[modif]+=1.0
                    else:
                        quarter2_4[modif]=1.0
                    counters4[1]+=1
                else:
                    if modif in quarter3_4:
                        quarter3_4[modif]+=1.0
                    else:
                        quarter3_4[modif]=1.0
                    counters4[2]+=1
        
        quarter1_4=self.__mb_refine(quarter1_4)
        quarter2_4=self.__mb_refine(quarter2_4)
        quarter3_4=self.__mb_refine(quarter3_4)
        keys4=list(set(quarter1_4.keys()+quarter2_4.keys()+quarter3_4.keys()))
        total4=total_counter4
        mean4={}
        for k in keys4:
            mean4[k] = [round(quarter1_4[k]/total4,5)] if k in quarter1_4 else [0.0]
            mean4[k].append(round(quarter2_4[k]/total4,5) if k in quarter2_4 else 0.0)
            mean4[k].append(round(quarter3_4[k]/total4,5) if k in quarter3_4 else 0.0)
        
        counters6=[0,0,0]
        quarter1_6={}
        quarter2_6={}
        quarter3_6={}
        for m,s in src_tab6[t]:
            for modif, hop in m.iteritems():
                if hop<=3:
                    if modif in quarter1_6:
                        quarter1_6[modif]+=1.0
                    else:
                        quarter1_6[modif]=1.0
                    counters6[0]+=1
                elif hop <=9:
                    if modif in quarter2_6:
                        quarter2_6[modif]+=1.0
                    else:
                        quarter2_6[modif]=1.0
                    counters6[1]+=1
                else:
                    if modif in quarter3_6:
                        quarter3_6[modif]+=1.0
                    else:
                        quarter3_6[modif]=1.0
                    counters6[2]+=1
        
        quarter1_6=self.__mb_refine(quarter1_6)
        quarter2_6=self.__mb_refine(quarter2_6)
        quarter3_6=self.__mb_refine(quarter3_6)
        keys6=list(set(quarter1_6.keys()+quarter2_6.keys()+quarter3_6.keys()))
        total6=total_counter6
        mean6={}
        for k in keys6:
            mean6[k] = [round(quarter1_6[k]/total6,5)] if k in quarter1_6 else [0.0]
            mean6[k].append(round(quarter2_6[k]/total6,5) if k in quarter2_6 else 0.0)
            mean6[k].append(round(quarter3_6[k]/total6,5) if k in quarter3_6 else 0.0)
       
        N4=3.0 # nb monitor
        cm=plt.get_cmap("Paired")
        label4=["Close to VP", "core","Close to Dst"]
        N6=3.0
        ind = np.arange(N4)  # the x locations for the groups
        width = 1.0/len(keys4)     # the width of the bars

        rects=[]
        fig, ax = plt.subplots()
        i=0
        for s in mean4:
            rects.append(ax.bar(ind+i*width, tuple(mean4[s]), width, color=cm(1.*i/len(keys4)),label=s))
            i+=1
        
        # add text
        ax.set_ylabel('Rate')
        ax.set_title('IPv4 commons middleboxes modifications')
        ax.legend(prop={'size':8}, loc=3,bbox_to_anchor=(1, 0),fancybox=True, shadow=True)

        ax.set_xticks(ind)
        ax.set_xticklabels(tuple(label4),rotation=40,ha='right')
        for v in ind:
            ax.plot([ind, ind],[0, 1], linewidth=2, linestyle='dotted', color='0.7')
        plt.yscale('log', nonposy='clip')
        plt.tight_layout()
        plt.subplots_adjust(right=0.7)
        plt.savefig(self.output_folder+'mb_quarters_overall_v4.pdf')
        plt.close(fig)
        del fig       
        
        cm=plt.get_cmap("Paired")
        width2 = 1.0/len(keys6)
        rects2=[]
        fig2, ax2 = plt.subplots()
        i=0
        for s in mean6:
            rects2.append(ax2.bar(ind+i*width2, tuple(mean6[s]), width2, color=cm(1.*i/len(keys6)),label=s))
            i+=1
        
        # add text
        ax2.set_ylabel('Rate')
        ax2.set_title('IPv6 commons middleboxes modifications')
        ax2.legend(prop={'size':8}, loc=3,bbox_to_anchor=(1, 0),fancybox=True, shadow=True)

        ax2.set_xticks(ind)
        ax2.set_xticklabels(tuple(label4),rotation=40,ha='right')
        for v in ind:
            ax2.plot([ind, ind],[0, 1], linewidth=2, linestyle='dotted', color='0.7')
        plt.yscale('log', nonposy='clip')
        plt.tight_layout()
        plt.subplots_adjust(right=0.65)
        plt.savefig(self.output_folder+'mb_quarters_overall_v6.pdf')
        plt.close(fig2)
        del fig2

        print "Plots outputed to <plot-dir>/modifs.log"
                    
    def icmp_size_stats(self):
        """
        Ouputs:
        total + per monitor
        total + per monitor / distinct hop
        each with distinct rate of each size class plus the average size
        """
        src_tab4={}
        total_size4={}
        src_tab6={}
        total_size6={}

        # Print summary
        print "Analyzing probes to"
        print str(len(self.parsebox.array4.keys()))+" IPv4 destinations."
        print str(len(self.parsebox.array6.keys()))+" IPv6 destinations."    
        print
        
        # IPv4
        path4_with=0.0
        path4_total=0.0
        for dst in self.parsebox.array4:
            for src in self.parsebox.array4[dst]:
                if "127.0.0.1" in src:
                    continue
                succ=0
                full=0
                for tb in self.parsebox.array4[dst][src]:
                    succ=1
                    for hop_addr,hop_icmpsize in tb.get_size_list():  
                        if hop_icmpsize == "(full)":
                            full=1
                            break            
                    if full:
                        break
                #for src
                if succ != 0:
                    path4_with+=full
                    path4_total+=1
                    if src in src_tab4:
                        e1,e2=src_tab4[src]
                        src_tab4[src]=[e1+full,e2+1]
                    else:
                        src_tab4[src]=[full,1]

        print "%d IPv4 paths contained at least 1 RFC1812 router among %d (%.2f%%)" % (path4_with,path4_total,(path4_with*100.0)/path4_total)
        for src in src_tab4:
            print "%s.%s: %d paths among %d contained >= 1 RFC1812 router (%.2f%%)" % (self.ip_to_label(src,4),self.ip_to_continent(src,4),src_tab4[src][0],src_tab4[src][1],(src_tab4[src][0]*100)/src_tab4[src][1])
     
        # IPv6
        path6_with=0.0
        path6_total=0.0
        for dst in self.parsebox.array6:
            for src in self.parsebox.array6[dst]:
                if "::1" in src:
                    continue
                succ=0
                full=0
                for tb in self.parsebox.array6[dst][src]:

                    succ=1
                    
                    for hop_addr,hop_icmpsize in tb.get_size_list():  
                        if hop_icmpsize == "(full)":
                            full=1
                            break            
                    if full:
                        break
                #for src
                if succ != 0:
                    path6_with+=full
                    path6_total+=1
                    if src in src_tab6:
                        e1,e2=src_tab6[src]
                        src_tab6[src]=[e1+full,e2+1]
                    else:
                        src_tab6[src]=[full,1]
        print "%d IPv6 paths contained at least 1 RFC1812 router among %d (%.2f%%)" % (path6_with,path6_total,(path6_with*100.0)/path6_total)
        for src in src_tab6:
            print "%s.%s: %d paths among %d contained >= 1 RFC1812 router (%.2f%%)" % (self.ip_to_label(src,6),self.ip_to_continent(src,6),src_tab6[src][0],src_tab6[src][1],(src_tab6[src][0]*100)/src_tab6[src][1])
        

    def icmp_size_plots(self):
        """
        Ouputs plots about quoting packets sizes.
        """
        src_tab4={}
        total_size4={}
        src_tab6={}
        total_size6={}
        for dst in self.parsebox.array4:
            for src in self.parsebox.array4[dst]:
                if "127.0.0.1" in src:
                    continue
                for tb in self.parsebox.array4[dst][src]:
                    for hop_addr,hop_icmpsize in tb.get_size_list():
                        
                        if hop_addr in total_size4:
                            if hop_icmpsize in total_size4[hop_addr]:
                                total_size4[hop_addr][hop_icmpsize]+=1.0
                            else:
                                total_size4[hop_addr][hop_icmpsize]=1.0
                        else:
                            total_size4[hop_addr]={}
                            total_size4[hop_addr][hop_icmpsize]=1.0
                                
                        if src in src_tab4:
                            if hop_addr in src_tab4[src]:
                                if hop_icmpsize in src_tab4[src][hop_addr]:
                                    src_tab4[src][hop_addr][hop_icmpsize]+=1.0
                                else:
                                    src_tab4[src][hop_addr][hop_icmpsize]=1.0
                            else:
                                src_tab4[src][hop_addr]={}
                                src_tab4[src][hop_addr][hop_icmpsize]=1.0
                        else:
                            src_tab4[src]={}
                            src_tab4[src][hop_addr]={}
                            src_tab4[src][hop_addr][hop_icmpsize]=1.0
                            
        src_tab6={}
        total_size6={}
        for dst in self.parsebox.array6:
            for src in self.parsebox.array6[dst]:
                if src == "::1":
                    continue
                for tb in self.parsebox.array6[dst][src]:
                    for hop_addr,hop_icmpsize in tb.get_size_list():
                        
                        if hop_addr in total_size6:
                            if hop_icmpsize in total_size6[hop_addr]:
                                total_size6[hop_addr][hop_icmpsize]+=1.0
                            else:
                                total_size6[hop_addr][hop_icmpsize]=1.0
                        else:
                            total_size6[hop_addr]={}
                            total_size6[hop_addr][hop_icmpsize]=1.0
                                
                        if src in src_tab6:
                            if hop_addr in src_tab6[src]:
                                if hop_icmpsize in src_tab6[src][hop_addr]:
                                    src_tab6[src][hop_addr][hop_icmpsize]+=1.0
                                else:
                                    src_tab6[src][hop_addr][hop_icmpsize]=1.0
                            else:
                                src_tab6[src][hop_addr]={}
                                src_tab6[src][hop_addr][hop_icmpsize]=1.0
                        else:
                            src_tab6[src]={}
                            src_tab6[src][hop_addr]={}
                            src_tab6[src][hop_addr][hop_icmpsize]=1.0

        # per hops all monitors together
        count_total_size4={k:sum(d.values()) for k,d in total_size4.iteritems()}
        count_total_size6={k:sum(d.values()) for k,d in total_size6.iteritems()}
        # to get the overall total
        overall_count_total_size4=sum(count_total_size4.values())
        overall_count_total_size6=sum(count_total_size6.values())
        # per hops per src
        count_src_tab4={k:{k2:sum(d2.values()) for k2,d2 in d.iteritems()} for k,d in src_tab4.iteritems()}
        count_src_tab6={k:{k2:sum(d2.values()) for k2,d2 in d.iteritems()} for k,d in src_tab6.iteritems()}
        #count per src, hops together
        overall_count_src_tab4={k:sum(d.values()) for k,d in count_src_tab4.iteritems()}
        overall_count_src_tab6={k:sum(d.values()) for k,d in count_src_tab6.iteritems()}
        
        #count occurences of each icmp size (total and per monitor)
        c=Counter()
        for k,d in total_size4.iteritems():
            c.update(d)
        mean4={k:[round(d/overall_count_total_size4,5)] for k,d in dict(c).iteritems()}
        
        label4=["total"]
        for k,d in src_tab4.iteritems():
            label4.append(self.ip_to_label(k,4))
            c.clear()
            for k2,d2 in d.iteritems():
                c.update(d2)
            dic=dict(c)
            for size in mean4:
                if size in dic:
                    mean4[size].append(round(dic[size]/overall_count_src_tab4[k],5))
                else:
                    mean4[size].append(0.0)
                    
        c.clear()
        for k,d in total_size6.iteritems():
            c.update(d)
        mean6={k:[round(d/overall_count_total_size6,5)] for k,d in dict(c).iteritems()}
        label6=["total"]
        for k,d in src_tab6.iteritems():
            label6.append(self.ip_to_label(k,6))
            c.clear()
            for k2,d2 in d.iteritems():
                c.update(d2)
            dic=dict(c)
            for size in mean6:
                if size in dic:
                    mean6[size].append(round(dic[size]/overall_count_src_tab6[k],5))
                else:
                    mean6[size].append(0.0)
        

        N4=len(src_tab4)+1.0 # nb monitor
        N6=len(src_tab6)+1.0
        ind = np.arange(N4)  # the x locations for the groups
        width = 1.0/len(mean4.keys())     # the width of the bars
        colors=['r','b','y','c']
        labels={'(0)':"empty", '(L3)': "layer 3 only", '(8B)' : "Layer 3 + 8B", '(full)': "Full segment"}
        rects=[]
        fig, ax = plt.subplots()
        i=0
        for s in mean4:
            rects.append(ax.bar(ind+i*width, tuple(mean4[s]), width,color=colors[i], label=labels[s]))
            i+=1
        
        # add text
        ax.set_ylabel('Rate')
        ax.set_title('IPv4 icmp data size')
        ax.legend(prop={'size':8}, loc=3,bbox_to_anchor=(1, 0),fancybox=True, shadow=True)
        #plt.subplots_adjust(bottom=0.2,right=0.8,top=0.8)

        ax.set_xticks(ind+(width/2))
        ax.set_xticklabels(tuple(label4),rotation=40,ha='right')
        
        plt.tight_layout()
        plt.subplots_adjust(right=0.75)
        plt.savefig(self.output_folder+'icmp_size_overall_v4.pdf')
        plt.close(fig)
        del fig
        
        ind2 = np.arange(N6)  # the x locations for the groups
        width2 = 1.0/len(mean6.keys())       # the width of the bars
        fig2, ax2 = plt.subplots()
        i=0
        for s in mean6:
            rects.append(ax2.bar(ind2+i*width2, tuple(mean6[s]), width2,color=colors[i], label=labels[s]))
            i+=1

        # add text
        ax2.set_ylabel('Rate')
        ax2.set_title('IPv6 icmp data size')
        #box2 = ax2.get_position()
        #ax2.set_position([box2.x0, box2.y0, box2.width * 0.8, box2.height])
        ax2.legend( prop={'size':8}, loc=3,bbox_to_anchor=(1, 0),fancybox=True, shadow=True)
    
        ax2.set_xticks(ind2+(width2/2))
        ax2.set_xticklabels(tuple(label6),rotation=40,ha='right')
        
        plt.tight_layout()
        #plt.subplots_adjust(right=0.65)
        plt.subplots_adjust(right=0.75)
        plt.savefig(self.output_folder+'icmp_size_overall_v6.pdf')
        plt.close(fig2)
        del fig2
        
        #transforming data to per-hop
        total_hop_count4=len(total_size4)
        hop_count4={k: len(d) for k,d in src_tab4.iteritems()}
        total_hop_size4={k: max(d.iteritems(), key=operator.itemgetter(1))[0] for k,d in total_size4.iteritems()}
        src_hop_size4={k: {k2: max(d2.iteritems(), key=operator.itemgetter(1))[0] for k2,d2 in d.iteritems()} for k,d in src_tab4.iteritems()}

        total_hop_count6=len(total_size6)
        hop_count6={k: len(d) for k,d in src_tab6.iteritems()}
        total_hop_size6={k: max(d.iteritems(), key=operator.itemgetter(1))[0] for k,d in total_size6.iteritems()}
        src_hop_size6={k: {k2: max(d2.iteritems(), key=operator.itemgetter(1))[0] for k2,d2 in d.iteritems()} for k,d in src_tab6.iteritems()}

        #count occurences of each icmp size (total and per monitor)
        mean4_2={k:[round(total_hop_size4.values().count(k)/float(total_hop_count4),5)] for k in mean4.keys()}
        label4=["total"]
        for src,d in src_hop_size4.iteritems():
            label4.append(self.ip_to_label(src,4))
            tmp={k:round(src_hop_size4[src].values().count(k)/float(hop_count4[src]),5) for k in mean4.keys()}
            for k in mean4.keys():
                if k in tmp:
                    mean4_2[k].append(tmp[k])
                else:
                    mean4_2[k].append(0.0)
                    
                #count occurences of each icmp size (total and per monitor)
        mean6_2={k:[round(total_hop_size6.values().count(k)/float(total_hop_count6),5)] for k in mean6.keys()}
        label6=["total"]
        for src,d in src_hop_size6.iteritems():
            label6.append(self.ip_to_label(src,6))
            tmp={k:round(src_hop_size6[src].values().count(k)/float(hop_count6[src]),5) for k in mean6.keys()}
            for k in mean6.keys():
                if k in tmp:
                    mean6_2[k].append(tmp[k])
                else:
                    mean6_2[k].append(0.0)
        
        # per-hop plots
        N4=len(src_tab4)+1.0 # nb monitor
        N6=len(src_tab6)+1.0

        # IPv4 per-hop
        ind = np.arange(N4)  # the x locations for the groups
        width = 1.0/len(mean4_2.keys())     # the width of the bars
        labels={'(0)':"empty", '(L3)': "layer 3 only", '(8B)' : "Layer 3 + 8B", '(full)': "Full segment"}
        rects=[]
        fig, ax = plt.subplots()
        i=0
        for s in mean4_2:
            rects.append(ax.bar(ind+i*width, tuple(mean4_2[s]), width,color=colors[i], label=labels[s]))
            i+=1
        
        # add text
        ax.set_ylabel('Rate')
        ax.set_title('IPv4 icmp data size')
        ax.legend(prop={'size':8}, loc=3,bbox_to_anchor=(1, 0),fancybox=True, shadow=True)
        #plt.subplots_adjust(bottom=0.2,right=0.8,top=0.8)

        ax.set_xticks(ind+(width/2))
        ax.set_xticklabels(tuple(label4),rotation=40,ha='right')
        
        plt.tight_layout()
        plt.subplots_adjust(right=0.75)
        plt.savefig(self.output_folder+'icmp_size_perhop_v4.pdf')
        plt.close(fig)
        del fig
        
        # IPv6 per-hop
        ind2 = np.arange(N6)  # the x locations for the groups
        width2 = 1.0/len(mean6.keys())       # the width of the bars
        fig2, ax2 = plt.subplots()
        i=0
        for s in mean6_2:
            rects.append(ax2.bar(ind2+i*width2, tuple(mean6_2[s]), width2,color=colors[i], label=labels[s]))
            i+=1
        
        # add text
        ax2.set_ylabel('Rate')
        ax2.set_title('IPv6 icmp data size')
        #box2 = ax2.get_position()
        #ax2.set_position([box2.x0, box2.y0, box2.width * 0.8, box2.height])
        ax2.legend( prop={'size':8}, loc=3,bbox_to_anchor=(1, 0),fancybox=True, shadow=True)
    
        ax2.set_xticks(ind2+(width2/2))
        ax2.set_xticklabels(tuple(label6),rotation=40,ha='right')
        
        plt.tight_layout()
        plt.subplots_adjust(right=0.75)
        plt.savefig(self.output_folder+'icmp_size_perhop_v6.pdf')
        plt.close(fig2)
        del fig2
        
        print "Plots outputed to <plot-dir>/modifs.log"
    
    def ip_to_continent(self,ip,v):
        """
        IP to continent lookup.
        """
        arr={}
        if v == 4:
            arr=self.monitors_v4
        elif v == 6:
            arr=self.monitors_v6
            
        if ip in arr:
            return arr[ip][1]
        else:
            return ip  
        
    def ip_to_label(self,ip,v):
        """
        IP to label lookup.
        """
        arr={}
        if v == 4:
            arr=self.monitors_v4
        elif v == 6:
            arr=self.monitors_v6
            
        if ip in arr:
            return arr[ip][0]
        else:
            return ip
     
    def success_rate(self):
        """
        Plots and print statistics about scamper states of tracebox runs.
        """
        # IPv4
        src_tab4={}
        total_status4={}
        for dst in self.parsebox.array4:
            for src in self.parsebox.array4[dst]:
                if "127.0.0.1" in src:
                    continue
                for tb in self.parsebox.array4[dst][src]:
                    if tb.status in total_status4:
                        total_status4[tb.status]+=1.0
                    else:
                        total_status4[tb.status]=1.0
                    if src in src_tab4:
                        if tb.status in src_tab4[src]:
                            src_tab4[src][tb.status]+=1.0
                        else:
                            src_tab4[src][tb.status]=1.0
                    else:
                        src_tab4[src]={}
                        src_tab4[src][tb.status]=1.0
        # IPv6
        src_tab6={}
        total_status6={}
        for dst in self.parsebox.array6:
            for src in self.parsebox.array6[dst]:
                if src == "::1":
                    continue
                for tb in self.parsebox.array6[dst][src]:
                    if tb.status in total_status6:
                        total_status6[tb.status]+=1.0
                    else:
                        total_status6[tb.status]=1.0
                    if src in src_tab6:
                        if tb.status in src_tab6[src]:
                            src_tab6[src][tb.status]+=1.0
                        else:
                            src_tab6[src][tb.status]=1.0
                    else:
                        src_tab6[src]={}
                        src_tab6[src][tb.status]=1.0
                        
        N4=len(src_tab4)+1.0 # nb monitor
        N6=len(src_tab6)+1.0
        total4=sum(total_status4.values())
        total6=sum(total_status6.values()) 
          
        mean4={}
        for s in total_status4:
            mean4[s]=[round(total_status4[s]/total4,5)]
        mean6={}
        for s in total_status6:
            mean6[s]=[round(total_status6[s]/total6,5)]

        print "IPv4 status in proportion of paths."
        print mean4
        print
        print "IPv6 status in proportion of paths."
        print mean6
        print

        label4=["total"]
        for src in src_tab4:
            label4.append(self.ip_to_label(src,4)) 
            total_src=sum(src_tab4[src].values())
            for status in mean4:
                if status in src_tab4[src]:
                    mean4[status].append(round(src_tab4[src][status]/total_src,5))
                else:
                    mean4[status].append(0.0)
                    
        label6=["total"]
        for src in src_tab6:
            label6.append(self.ip_to_label(src,6))
            total_src=sum(src_tab6[src].values())
            for status in mean6:
                if status in src_tab6[src]:
                    mean6[status].append(round(src_tab6[src][status]/total_src,5))
                else:
                    mean6[status].append(0.0)

        
        ind = np.arange(N4)  # the x locations for the groups
        width = 1.0/len(mean4.keys())      # the width of the bars
        colors=['r','b','y','c','m']
        rects=[]
        fig, ax = plt.subplots()
        i=0
        for s in mean4:
            rects.append(ax.bar(ind+i*width, tuple(mean4[s]), width,color=colors[i], label=s))
            i+=1
            
        # add text
        ax.set_ylabel('Rate')
        ax.set_title('IPv4 probe status')
        ax.legend(prop={'size':8}, loc=3,bbox_to_anchor=(1, 0),fancybox=True, shadow=True)
        #plt.subplots_adjust(bottom=0.2,right=0.8,top=0.8)

        ax.set_xticks(ind+(width/2))
        ax.set_xticklabels(tuple(label4),rotation=40,ha='right')
        
        plt.tight_layout()
        plt.subplots_adjust(right=0.75)
        plt.savefig(self.output_folder+'successv4.pdf')
        plt.close(fig)
        del fig
        
        ind2 = np.arange(N6)  # the x locations for the groups
        width2 = 1.0/len(mean6.keys())       # the width of the bars
        fig2, ax2 = plt.subplots()
        i=0
        for s in mean6:
            rects.append(ax2.bar(ind2+i*width2, tuple(mean6[s]), width2,color=colors[i], label=s))
            i+=1
            
        # add text
        ax2.set_ylabel('Rate')
        ax2.set_title('IPv6 probe status')
        #box2 = ax2.get_position()
        #ax2.set_position([box2.x0, box2.y0, box2.width * 0.8, box2.height])
        ax2.legend( prop={'size':8}, loc=3,bbox_to_anchor=(1, 0),fancybox=True, shadow=True)
    
        ax2.set_xticks(ind2+(width2/2))
        ax2.set_xticklabels(tuple(label6),rotation=40,ha='right')
        
        plt.tight_layout()
        #plt.subplots_adjust(right=0.65)
        plt.subplots_adjust(right=0.75)
        plt.savefig(self.output_folder+'successv6.pdf')
        plt.close(fig2)
        del fig2
    
    def rtts_diff(self):
        """
        compute RTT differences
        """
        t="total"
        src_tab4={t:[]}
        total_counter4=0
        for dst in self.parsebox.array4:
            for src in self.parsebox.array4[dst]:
                if "127.0.0.1" in src:
                    continue
                for tb in self.parsebox.array4[dst][src]:
                    if tb.status != "success":
                        continue
                    total_counter4+=1
                    rtt = tb.get_rtt()
                    continent=self.ip_to_continent(src, 4)
                    src_tab4[t].append(rtt)
                    if continent in src_tab4:
                        src_tab4[continent].append(rtt)
                    else:
                        src_tab4[continent]=[rtt]
                   
        diff_tab={t:[]}     
        src_tab6={t:[]}
        total_counter6=0
        for dst in self.parsebox.array6:
            for src in self.parsebox.array6[dst]:
                if src == "::1":
                    continue
                for tb in self.parsebox.array6[dst][src]:
                    if tb.status != "success":
                        continue
                    total_counter6+=1
                    rtt6 = tb.get_rtt()
                    dst4=self.parsebox.get_v4_from_v6(dst)
                    src4=self.parsebox.get_monitor_v4_from_v6(src)
                    rtt4=None
                    
                    if dst4 in self.parsebox.array4:
                        if src4 in self.parsebox.array4[dst4]:
                            for tb in self.parsebox.array4[dst4][src4]:
                                if tb.status != "success":
                                    continue
                                rtt4 = tb.get_rtt()
                                break
                            
                    continent=self.ip_to_continent(src, 6)
                    if rtt4 and rtt6:
                        if rtt4==0.0 or rtt6 ==0.0:
                            val=0.0
                        else:
                            val=(rtt4-rtt6)/rtt4
                        diff_tab[t].append(val)
                        
                        if continent in diff_tab:
                            diff_tab[continent].append(val)
                        else:
                            diff_tab[continent]=[val]
                            
                    src_tab6[t].append(rtt6)
                    if continent in src_tab6:
                        src_tab6[continent].append(rtt6)
                    else:
                        src_tab6[continent]=[rtt6]
        mean_rtt={}
        std_rtt={}
        n4={}
        for continent in diff_tab:
            sample4 = np.array(diff_tab[continent])
            mean_rtt[continent]=sample4.mean()
            std_rtt[continent]=sample4.std()
            fig, ax = plt.subplots()
            n,bins,patches=ax.hist(sample4, 200, range=(-5.0,5.0), normed=1, histtype='bar', color='g', alpha=0.8)
            n4[continent]=n
            
            ax.set_ylabel('Percentage of probes to all destinations')
            ax.set_xlabel('Relative RTT differences')
            ax.set_title('v4/v6  RTTs Relative difference')
            plt.tight_layout()
            plt.savefig(self.output_folder+'rtt_hist_cmp_'+continent+'.pdf')
            plt.close(fig)
            del fig
        
            clean=[]
            for v in diff_tab[continent]:
                if v>-5.0 and v<5.0:
                    clean.append(v)
            sample4=np.array(clean)
            ecdf = sm.distributions.ECDF(sample4)
            x = np.linspace(min(sample4), max(sample4))
            y = ecdf(x)
            xmin=min(sample4)
            xmax=max(sample4)
            fig = plt.figure()
            #fig.suptitle('bold figure suptitle', fontsize=14, fontweight='bold')
            ax = fig.add_subplot(111)
            #ax.set_title('axes title')
            ax.set_xlabel('Relative RTT differences')
            ax.set_ylabel('Cumulative distribution function (CDF)')
            ax.set_title('v4/v6 RTTs comparison')
            ax.annotate("Longer IPv4\n RTT",(xmax*.2,0.6))
            ax.annotate("Shorter IPv4\n RTT",(xmin*.75,0.6))
  
            ax.plot([0.0, 0.0],[0, 1], linewidth=2, linestyle='dotted', color='0.7')
            ax.step(x, y)
            ax.legend()
            plt.tight_layout()
            plt.savefig(self.output_folder+'rtt_rel_diff_cmp_'+continent+'.pdf')    
            plt.close(fig)
            del fig

        # write means, std dev
        f=open(self.output_folder+"rtts_cmp.log",'w')
        f.write("(rtt4-rtt6)/rtt4:\n\n")
        for continent in mean_rtt:
            f.write("%s\n"%continent)
            f.write("Mean: %.2f \n"%mean_rtt[continent])
            f.write("Standard deviation: %.2f \n"%std_rtt[continent])
            f.write("\n")
        f.close()         
    
    def rtts(self):
        t="total"
        src_tab4={t:[]}
        total_counter4=0
        for dst in self.parsebox.array4:
            for src in self.parsebox.array4[dst]:
                if "127.0.0.1" in src:
                    continue
                for tb in self.parsebox.array4[dst][src]:
                    if tb.status != "success":
                        continue
                    total_counter4+=1
                    rtt = tb.get_rtt()
                    continent=self.ip_to_continent(src, 4)
                    src_tab4[t].append(rtt)
                    if continent in src_tab4:
                        src_tab4[continent].append(rtt)
                    else:
                        src_tab4[continent]=[rtt]
                        
        src_tab6={t:[]}
        total_counter6=0
        for dst in self.parsebox.array6:
            for src in self.parsebox.array6[dst]:
                if src == "::1":
                    continue
                for tb in self.parsebox.array6[dst][src]:
                    if tb.status != "success":
                        continue
                    total_counter6+=1
                    rtt = tb.get_rtt()
                    continent=self.ip_to_continent(src, 6)
                    src_tab6[t].append(rtt)
                    if continent in src_tab6:
                        src_tab6[continent].append(rtt)
                    else:
                        src_tab6[continent]=[rtt]
        
        mean_rtt4={}
        std_rtt4={}
        n4={}
        for continent in src_tab4:
            sample4 = np.array(src_tab4[continent])
            mean_rtt4[continent]=sample4.mean()
            std_rtt4[continent]=sample4.std()
            fig, ax = plt.subplots()
            n,bins,patches=ax.hist(sample4, 250, range=(0,500), normed=1, histtype='bar', color='g', rwidth=0.8, alpha=0.8)
            #y = mlab.normpdf(bins, mean_rtt4[continent], std_rtt4[continent])
            n4[continent]=n
            
            ax.set_ylabel('Probes to all destinations (%)')
            ax.set_xlabel('RTTs in milliseconds')
            #ax.set_title('IPv4 RTTs ')
            #plt.plot(bins, y, 'r--')
            plt.tight_layout()
            plt.savefig(self.output_folder+'rtt_hist_v4_'+continent+'.pdf')
            plt.close(fig)
            del fig
            
        mean_rtt6={}
        std_rtt6={}
        n6={}
        for continent in src_tab6:
            sample6 = np.array(src_tab6[continent])
            mean_rtt6[continent]=sample6.mean()
            std_rtt6[continent]=sample6.std()
            fig, ax = plt.subplots()
            n,bins,patches=ax.hist(sample6, 250, range=(0,500), normed=1, histtype='bar', color='g', rwidth=0.8, alpha=0.8)
            #y = mlab.normpdf(bins, mean_rtt4[continent], std_rtt4[continent])
            n6[continent]=n
            
            ax.set_ylabel('Probes to all destinations (%)')
            ax.set_xlabel('RTTs in milliseconds')
            #ax.set_title('IPv6 RTTs ')
            #plt.plot(bins, y, 'r--')
            plt.tight_layout()
            plt.savefig(self.output_folder+'rtt_hist_v6_'+continent+'.pdf')
            plt.close(fig)
            del fig
            
            
        # write means, std dev
        f=open(self.output_folder+"rtts.log",'w')
        f.write("IPv4:\n\n")
        for continent in mean_rtt4:
            f.write("%s\n"%continent)
            f.write("Mean: %.2f \n"%mean_rtt4[continent])
            f.write("Standard deviation: %.2f \n"%std_rtt4[continent])
            f.write("\n")
        f.write("IPv6:\n\n")
        for continent in mean_rtt6:
            f.write("%s\n"%continent)
            f.write("Mean: %.2f \n"%mean_rtt6[continent])
            f.write("Standard deviation: %.2f \n"%std_rtt6[continent])
            f.write("\n")
        f.close()        
        
        for continent in n6:   
            nn4=n4[continent]
            nn6=n6[continent]
            fig, ax4 = plt.subplots()
            reldiff=[]
            for e4,e6 in zip(nn4,nn6):
                if e4 == 0.0:
                    reldiff.append(0.0)
                else:
                    reldiff.append((e4-e6)/e4)
            ax4.bar(bins[:-1], reldiff, color='r',alpha=0.7)
            ax4.set_ylabel('Relative RTTs Rate difference')
            ax4.set_xlabel('RTTs in milliseconds')
            ax4.set_title('v4/v6 RTTs comparison')
            ax4.annotate("Shorter IPv4 RTT",(200,min(reldiff)*.75))
            ax4.annotate("Longer IPv4 RTT",(200,max(reldiff)*.75))
            plt.tight_layout()
            plt.savefig(self.output_folder+'rtt_rel_diff_'+continent+'.pdf')
            plt.close(fig)
            del fig
            
    def path_hist_cmp(self,v=None,src_list=None,dst_list=None):
        t="total"
        diff_tab={t:[]}     
        total_counter6=0
        for dst in self.parsebox.array6:
            for src in self.parsebox.array6[dst]:
                if src == "::1":
                    continue
                for tb in self.parsebox.array6[dst][src]:
                    if tb.status != "success":
                        continue
                    total_counter6+=1
                    len6=None
                    len6 = float(tb.path_length())
                    dst4=self.parsebox.get_v4_from_v6(dst)
                    src4=self.parsebox.get_monitor_v4_from_v6(src)
                    len4=None
                    
                    if dst4 in self.parsebox.array4:
                        if src4 in self.parsebox.array4[dst4]:
                            for tb in self.parsebox.array4[dst4][src4]:
                                if tb.status != "success":
                                    continue
                                len4 = float(tb.path_length())
                                break
                            
                    continent=self.ip_to_continent(src, 6)
                    if len4 != None and len6 != None:
                        if len4==0.0 or len6 ==0.0:
                            val=0.0
                        else:
                            val=(len4-len6)/len4
                            diff_tab[t].append(val)
                        
                            if continent in diff_tab:
                                diff_tab[continent].append(val)
                            else:
                                diff_tab[continent]=[val]
        f=open(self.output_folder+"hop_count_cmp.log",'w')
        for continent in diff_tab:                     
            l6=diff_tab[continent]            
            sample6 = np.array(l6)
            mean6=sample6.mean()
            std6=sample6.std()
            # write means, std dev
            f.write("%s\n"%continent)
            f.write("Mean: %.2f \n"%mean6)
            f.write("Standard deviation: %.2f \n\n"%std6)
            
        f.close()
        
    def path_hist(self,v=None,src_list=None,dst_list=None):
        l6=self.__path_length_list(self.parsebox.array6)
        sample6 = np.array(l6)
        mean6=sample6.mean()
        std6=sample6.std()
        fig, ax = plt.subplots()
        bins=[i for i in range(0,37)]
        ax.hist(sample6, bins,  normed=1, histtype='bar', color='g', rwidth=0.8, alpha=0.8)
        y = mlab.normpdf(bins, mean6, std6)
        
        ax.set_ylabel('Probability density for all tested paths')
        ax.set_xlabel('Number of hops')
        ax.set_title('IPv6 hop count distribution ')
        plt.plot(bins, y, 'r--')
        plt.tight_layout()
        plt.savefig(self.output_folder+'hop_count_hist_v6.pdf')
        plt.close(fig)
        del fig
        
        
        l4=self.__path_length_list(self.parsebox.array4)
        sample4 = np.array(l4)
        mean4=sample4.mean()
        std4=sample4.std()
        fig, ax2 = plt.subplots()
        ax2.hist(sample4, bins, normed=1, histtype='bar',color='b', rwidth=0.8, alpha=0.8)
        y = mlab.normpdf(bins, mean4, std4)
        plt.plot(bins, y, 'r--')
        # add text
        ax2.set_ylabel('Probability density for all tested paths')
        ax2.set_xlabel('Number of hops')
        ax2.set_title('IPv4 hop count distribution ')
        plt.tight_layout()
        plt.savefig(self.output_folder+'hop_count_hist_v4.pdf')
        plt.close(fig)
        del fig
        
        
        fig, ax3 = plt.subplots()
        n1, bins1, patches1 =ax3.hist(sample4, bins, normed=1, stacked=1, alpha=0.5,label="IPv4 paths")
        n2, bins2, patches2 =ax3.hist(sample6, bins, normed=1, stacked=1,alpha=0.5,label="IPv6 paths")
        
        ax3.set_ylabel('Probability density for all tested paths')
        ax3.set_xlabel('Number of hops')
        ax3.set_title('Hop count distribution ')
        ax3.legend(prop={'size':8}, loc=3,bbox_to_anchor=(1, 0),fancybox=True, shadow=True)
        #plt.subplots_adjust(bottom=0.2,right=0.8,top=0.8)
        plt.tight_layout()
        plt.subplots_adjust(right=0.75)
        plt.savefig(self.output_folder+'hop_count_hist.pdf')
        plt.close(fig)
        del fig
        
        # write means, std dev
        f=open(self.output_folder+"hop_count.log",'w')
        f.write("IPv4:\n\n")
        f.write("Mean: %.2f \n"%mean4)
        f.write("Standard deviation: %.2f \n"%std4)
        f.write("IPv6:\n\n")
        f.write("Mean: %.2f \n"%mean6)
        f.write("Standard deviation: %.2f \n"%std6)
        f.close()
        
        fig, ax4 = plt.subplots()
        reldiff=[]
        for e4,e6 in zip(n1,n2):
            reldiff.append(e4-e6)
        ax4.bar(bins1[:-2], reldiff[:-1], color='r',alpha=0.7)
        ax4.set_ylabel('Absolute Hop count Rate difference')
        ax4.set_xlabel('Number of hops')
        ax4.set_title('Hop count v4/v6 comparison')
        ax4.annotate("More IPv6 paths",(19,-0.03))
        ax4.annotate("More IPv4 paths",(19,0.04))
        plt.tight_layout()
        plt.savefig(self.output_folder+'hop_count_rel_diff.pdf')
        plt.close(fig)
        del fig
        
    def path_cdf_cmp(self,v=None,src_list=None,dst_list=None,minhops=2.0,maxhops=30.0):
        """
        plot path length cdf

        @param minhops the minimum path length
        @param maxhops the maximum path length
        """

        t="total"
        diff_tab={t:[]}     
        total_counter6=0
        lv4=[]
        lv6=[]
        for dst in self.parsebox.array6:
            for src in self.parsebox.array6[dst]:
                if src == "::1":
                    continue
                for tb in self.parsebox.array6[dst][src]:
                    if tb.status != "success":
                        continue
                    total_counter6+=1
                    len6=None
                    len6 = float(tb.path_length())
                    dst4=self.parsebox.get_v4_from_v6(dst)
                    src4=self.parsebox.get_monitor_v4_from_v6(src)
                    len4=None
                    
                    if dst4 in self.parsebox.array4:
                        if src4 in self.parsebox.array4[dst4]:
                            for tb in self.parsebox.array4[dst4][src4]:
                                if tb.status != "success":
                                    continue
                                len4 = float(tb.path_length())
                                break
                            
                    continent=self.ip_to_continent(src, 6)
                    if len4 != None and len6 != None:
                        if len4<=minhops or len6 <=minhops or len4>=maxhops or len6>=maxhops:
                            val=0.0
                        else:
                            # compute relative difference
                            val=(len4-len6)/len4
                            if val>-0.5 and val <= 0.5:
                                lv4.append(len4)
                                lv6.append(len6)
                                diff_tab[t].append(val)
                            
                                if continent in diff_tab:
                                    diff_tab[continent].append(val)
                                else:
                                    diff_tab[continent]=[val]
        arlv4=np.array(lv4)
        arlv6=np.array(lv6)
        print "Total IPv4 amount of paths."
        print len(arlv4)
        print "IPv4 hop number mean."
        print arlv4.mean()
        print "Total IPv6 mount of paths."
        print len(arlv6)
        print "IPv6 hop number mean."
        print arlv6.mean()
        print

        print "Per-continent amount of paths and hop number v4/v6 relative difference."
        for continent in diff_tab:
            l=diff_tab[continent]
            sample = np.array(l)
            print continent
            print sample.mean()
            print len(sample)
            ecdf = sm.distributions.ECDF(sample)
            xmin=min(sample)
            xmax=max(sample)
            x = np.linspace(xmin, xmax)
            y = ecdf(x)
            
            fig = plt.figure()
            ax = fig.add_subplot(111)
            ax.set_xlabel('Path length')
            ax.set_ylabel('Cumulative distribution function (CDF)')
            ax.plot([0, 0],[0, 1], linewidth=2, linestyle='dotted', color='0.7')
            ax.annotate("Shorter IPv4\n path",(xmin*.75,.5))
            ax.annotate("Longer IPv4\n path",(xmax*.25,.5))
            ax.step(x, y)
            ax.legend()
    
            if self.output_folder == None:
                plt.show()
            else:
                plt.savefig(self.output_folder+'cmp_'+continent+'_cdf.pdf')
                
            plt.close(fig)
            del fig      
    
    def path_cdf(self,v=None,src_list=None,dst_list=None):
        """
        plot path length cdf
        default output v4 and v6
        """
        
        l=self.__path_length_list(self.parsebox.array6)
        sample = np.array(l)
        ecdf = sm.distributions.ECDF(sample)
        xmin=min(sample)
        xmax=max(sample)
        x = np.linspace(xmin, xmax)
        y = ecdf(x)
        print "IPv6 path length mean."
        print sample.mean()
        print str(len(sample))+" IPv6 paths tested."
        print 
        
        l=self.__path_length_list(self.parsebox.array4)
        sample = np.array(l)
        ecdf = sm.distributions.ECDF(sample)
        x2 = np.linspace(min(sample), max(sample))
        y2 = ecdf(x2)
        print "IPv4 path length mean."
        print sample.mean()
        print str(len(sample))+" IPv6 paths tested."
        print 
        
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.set_xlabel('Path length')
        ax.set_ylabel('Cumulative distribution function')
        
        for i in range(5,100,5):
            if i>=xmax:
                break
            if i>=xmin:
                ax.plot([i, i],[0, 1], linewidth=2, linestyle='dotted', color='0.7')
        ax.step(x, y, label='IPv6')
        ax.step(x2, y2, label='IPv4')
        ax.legend()

        plt.tight_layout()
        if self.output_folder == None:
            plt.show()
        else:
            plt.savefig(self.output_folder+'cdf.pdf')
        
        plt.close(fig)
        del fig
        
